package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.citygml4j.core.model.core.AbstractSpaceBoundaryProperty;
import org.citygml4j.core.model.tunnel.TunnelInstallation;
import org.citygml4j.core.model.tunnel.TunnelInstallationProperty;
import org.citygml4j.core.util.geometry.GeometryFactory;

import java.io.Serial;
import java.util.ArrayList;
import java.util.List;

public class Tunnel extends AbstractTunnel{

    @Serial
    private static final long serialVersionUID = 5316281216192555555L;

    private List<TunnelPart> tunnelParts = new ArrayList<>();

    public List<TunnelPart> getTunnelParts() {
        return tunnelParts;
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        super.reCreateGeometries(factory, config);
        for (TunnelPart tp : tunnelParts) {
            tp.reCreateGeometries(factory, config);
        }
    }

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (TunnelPart tp : tunnelParts) {
            tp.accept(c);
        }
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (TunnelPart tp : tunnelParts) {
            tp.clearAllContainedCheckResults();
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (TunnelPart tp : tunnelParts) {
            tp.collectContainedErrors(errors);
        }
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (TunnelPart tp : tunnelParts) {
            if (tp.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (TunnelPart tp : tunnelParts) {
            if (tp.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    public void addTunnelPart(TunnelPart tunnelPart) {
        tunnelParts.add(tunnelPart);
    }

    @Override
    public String toString() {
        return "Tunnel [id=" + getGmlId() + "]";
    }

    public void anonymize() {
        for (Geometry geom : getGeometries()) {
            geom.anonymize();
        }
        org.citygml4j.core.model.tunnel.Tunnel gmlT = new org.citygml4j.core.model.tunnel.Tunnel();
        gmlT.setId(GmlId.generateId().getGmlString());
        for (Installation ti : getTunnelInstallations()) {
            ti.anonymize();
            gmlT.getTunnelInstallations().add(new TunnelInstallationProperty((TunnelInstallation) ti.getGmlObject()));
        }
        for (BoundarySurface bs : getBoundarySurfaces()) {
            bs.anonymize();
            gmlT.addBoundary(new AbstractSpaceBoundaryProperty(bs.getGmlObject()));
        }
        setCityGmlBuilding(gmlT);
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (TunnelPart tp : tunnelParts) {
            tp.clearMetaInformation();
        }
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (TunnelPart tp : tunnelParts) {
            tp.prepareForChecking();
        }
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        Tunnel originalTunnel = (Tunnel) original;
        for (TunnelPart originalTp : originalTunnel.tunnelParts) {
            tunnelParts.add(handler.getCopyInstance(originalTp));
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        for (TunnelPart tp : tunnelParts) {
            handler.addInstance(tp);
        }
    }

    @Override
    public Copyable createCopyInstance() {
        return new Tunnel();
    }


}
