/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.datastructure;

import java.io.Serial;
import java.util.Collections;

import org.citygml4j.core.model.core.AbstractCityObject;
import org.citygml4j.core.model.vegetation.AbstractVegetationObject;
import org.citygml4j.core.model.vegetation.PlantCover;
import org.citygml4j.core.model.vegetation.SolitaryVegetationObject;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSolid;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSolidProperty;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;
import org.xmlobjects.gml.model.geometry.primitives.Solid;
import org.xmlobjects.gml.model.geometry.primitives.SolidProperty;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.Copyable;

/**
 * Representation of CityGML vegetation objects
 * 
 * @author Matthias Betz
 *
 */
public class Vegetation extends CityObject {

	@Serial
	private static final long serialVersionUID = -5136358065541704146L;

	public enum VegetationType {
		SOLITARY_VEGETATION_OBJECT, PLANT_COVER
	}

	private AbstractVegetationObject citygmlVegetation;
	private final VegetationType type;

	public Vegetation(VegetationType type) {
		this.type = type;
	}

	@Override
	public FeatureType getFeatureType() {
		return FeatureType.VEGETATION;
	}

	@Override
	public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
		for (Geometry geom : getGeometries()) {
			if (geom instanceof ImplicitGeometryHolder){
				continue;
			}
			if (geom.getType() == GeometryType.MULTI_SURFACE) {
				MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
				if (type == VegetationType.SOLITARY_VEGETATION_OBJECT) {
					SolitaryVegetationObject svo = (SolitaryVegetationObject) citygmlVegetation;
					setMultiSurfaceAccordingToLod(svo, ms, geom.getLod());
				} else {
					PlantCover pc = (PlantCover) citygmlVegetation;
					setMultiSurfaceAccordingToLod(pc, ms, geom.getLod());
				}
			} else {
				Solid solid = CityGmlUtils.createSolid(geom, factory, config);
				if (type == VegetationType.SOLITARY_VEGETATION_OBJECT) {
					SolitaryVegetationObject svo = (SolitaryVegetationObject) citygmlVegetation;
					setSolidAccordingToLod(svo, solid, geom.getLod());
				} else {
					PlantCover pc = (PlantCover) citygmlVegetation;
					setSolidAccordingToLod(pc, solid, geom.getLod());
				}
			}
		}
	}

	private void setSolidAccordingToLod(PlantCover pc, Solid solid, Lod lod) {
		switch (lod) {
		case LOD1:
			pc.getDeprecatedProperties().setLod1MultiSolid(new MultiSolidProperty(new MultiSolid(Collections.singletonList(new SolidProperty(solid)))));
			break;
		case LOD2:
			pc.getDeprecatedProperties().setLod2MultiSolid(new MultiSolidProperty(new MultiSolid(Collections.singletonList(new SolidProperty(solid)))));
			break;
		case LOD3:
			pc.getDeprecatedProperties().setLod3MultiSolid(new MultiSolidProperty(new MultiSolid(Collections.singletonList(new SolidProperty(solid)))));
			break;
		case LOD4:
			pc.getDeprecatedProperties().setLod4MultiSolid(new MultiSolidProperty(new MultiSolid(Collections.singletonList(new SolidProperty(solid)))));
			break;
		default:
			throw new IllegalStateException("Cannot set Solid with lod to PlantCover:" + lod);
		}
	}

	private void setSolidAccordingToLod(SolitaryVegetationObject svo, Solid solid, Lod lod) {
		switch (lod) {
		case LOD1:
			svo.getDeprecatedProperties().setLod1Geometry(new SolidProperty(solid));
			break;
		case LOD2:
			svo.getDeprecatedProperties().setLod2Geometry(new SolidProperty(solid));
			break;
		case LOD3:
			svo.getDeprecatedProperties().setLod3Geometry(new SolidProperty(solid));
			break;
		case LOD4:
			svo.getDeprecatedProperties().setLod4Geometry(new SolidProperty(solid));
			break;
		default:
			throw new IllegalStateException("Cannot set Solid with lod to SolitaryVegetationObject:" + lod);
		}
	}

	private void setMultiSurfaceAccordingToLod(PlantCover pc, MultiSurface ms, Lod lod) {
		switch (lod) {
		case LOD1:
			pc.getDeprecatedProperties().setLod1MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD2:
			pc.setLod2MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD3:
			pc.setLod3MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD4:
			pc.getDeprecatedProperties().setLod4MultiSurface(new MultiSurfaceProperty(ms));
			break;
		default:
			throw new IllegalStateException("Cannot set MultiSurface with lod to PlantCover:" + lod);
		}

	}

	private void setMultiSurfaceAccordingToLod(SolitaryVegetationObject svo, MultiSurface ms, Lod lod) {
		switch (lod) {
		case LOD1:
			svo.getDeprecatedProperties().setLod1Geometry(new MultiSurfaceProperty(ms));
			break;
		case LOD2:
			svo.getDeprecatedProperties().setLod2Geometry(new MultiSurfaceProperty(ms));
			break;
		case LOD3:
			svo.getDeprecatedProperties().setLod3Geometry(new MultiSurfaceProperty(ms));
			break;
		case LOD4:
			svo.getDeprecatedProperties().setLod4Geometry(new MultiSurfaceProperty(ms));
			break;
		default:
			throw new IllegalStateException("Cannot set MultiSurface with lod to SolitaryVegetationObject:" + lod);
		}
	}

	@Override
	public void accept(Check c) {
		super.accept(c);
		if (c.canExecute(this)) {
			c.check(this);
		}
	}

	@Override
	public AbstractCityObject getGmlObject() {
		return citygmlVegetation;
	}

	@Override
	public void unsetGmlGeometries() {
		if (type == VegetationType.SOLITARY_VEGETATION_OBJECT) {
			SolitaryVegetationObject svo = (SolitaryVegetationObject) citygmlVegetation;
			svo.getDeprecatedProperties().setLod1Geometry(null);
			svo.getDeprecatedProperties().setLod2Geometry(null);
			svo.getDeprecatedProperties().setLod3Geometry(null);
			svo.getDeprecatedProperties().setLod4Geometry(null);
		} else {
			PlantCover pc = (PlantCover) citygmlVegetation;
			pc.getDeprecatedProperties().setLod1MultiSurface(null);
			pc.setLod2MultiSurface(null);
			pc.setLod3MultiSurface(null);
			pc.getDeprecatedProperties().setLod4MultiSurface(null);
		}
	}

	public void setGmlObject(AbstractVegetationObject avo) {
		citygmlVegetation = avo;
	}

	public VegetationType getVegetationType() {
		return type;
	}

	@Override
	public String toString() {
		return "Vegetation [id=" + getGmlId() + "]";
	}

	@Override
	public Copyable createCopyInstance() {
		return new Vegetation(type);
	}
}
